/*
 * bitmap.c - PM VNC Viewer, Bitmap (as backing store) Handlings
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#define INCL_PM
#include <os2.h>

#include "pmvncdef.h"

/*
 * bitmap for backing store
 */

HDC     hdcBitmap = NULLHANDLE ;
HPS     hpsBitmap = NULLHANDLE ;
HBITMAP hbmBitmap = NULLHANDLE ;

static  SIZEL   sizBitmap = { 0 } ;     /* size of RFB, fixed   */
static  SIZEL   sizClient = { 0 } ;     /* size of client area  */
static  POINTL  offClient = { 0 } ;     /* position on bitmap   */

/*
 * bmpDispose - cleanup bitmap resources
 */

void    bmpDispose(void)
{
    if (hbmBitmap != NULLHANDLE) {
        GpiDeleteBitmap(hbmBitmap) ;
	hbmBitmap = NULLHANDLE ;
    }
    if (hpsBitmap != NULLHANDLE) {
        GpiDestroyPS(hpsBitmap) ;
	hpsBitmap = NULLHANDLE ;
    }
    if (hdcBitmap != NULLHANDLE) {
        DevCloseDC(hdcBitmap) ;
	hdcBitmap = NULLHANDLE ;
    }
}

/*
 * bmpCreate - create & setup bitmap as backing store
 */

static  UCHAR   greeting[] = "Please wait - initial screen loading" ;

BOOL    bmpCreate(HAB hab, PSIZEL sizrfb)
{
    SIZEL   siz ;
    POINTL  pt  ;
    BITMAPINFOHEADER2   bmi ;
    
    /*
     * create bitmap of remote screen size
     */
     
    hdcBitmap = DevOpenDC(hab, OD_MEMORY, "*", 0, NULL, NULLHANDLE) ;
    siz.cx = siz.cy = 0 ;
    hpsBitmap = GpiCreatePS(hab, hdcBitmap, &siz,
                PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC) ;
    if (hdcBitmap == NULLHANDLE || hpsBitmap == NULLHANDLE) {
        bmpDispose() ;
	return FALSE ;
    }
    
    memset(&bmi, 0, sizeof(bmi)) ;
    bmi.cbFix = sizeof(bmi) ;
    bmi.cx = sizrfb->cx ;
    bmi.cy = sizrfb->cy ;
    bmi.cPlanes       = 1  ;
    bmi.cBitCount     = 24 ;
    bmi.ulCompression = 0  ;
    bmi.cclrUsed      = 0  ;
    bmi.cclrImportant = 0  ;

    hbmBitmap = GpiCreateBitmap(hpsBitmap, &bmi, 0, NULL, NULL) ;
    GpiSetBitmap(hpsBitmap, hbmBitmap) ;

    if (GpiCreateLogColorTable(hpsBitmap, 0, LCOLF_RGB, 0, 0, NULL) != TRUE) {
        TRACE("bmpCreate - failed to set color mode RGB\n") ;
	return FALSE ;
    }
    
    GpiErase(hpsBitmap) ;

    sizBitmap.cx = sizrfb->cx ;
    sizBitmap.cy = sizrfb->cy ;
    sizClient.cx = 0 ;
    sizClient.cy = 0 ;
    offClient.x  = 0 ;
    offClient.y  = sizBitmap.cy ;
    
    pt.x = 32 ;
    pt.y = sizBitmap.cy - 64 ;
    GpiCharStringAt(hpsBitmap, &pt, strlen(greeting), greeting) ;
    
    return TRUE ;
}

/*
 * bmpMapSize - adjust to client size
 */

void    bmpMapSize(SHORT cx, SHORT cy)
{
    SHORT   diff, vpos ;
    
    if ((offClient.y += (sizClient.cy - cy)) < 0) {
        offClient.y = 0 ;
    }
    sizClient.cx = (LONG) cx ;
    sizClient.cy = (LONG) cy ;
    
    if ((offClient.x + sizClient.cx) >= sizBitmap.cx) {
        offClient.x = sizBitmap.cx - sizClient.cx ;
    }
    if ((offClient.y + sizClient.cy) >= sizBitmap.cy) {
        offClient.y = sizBitmap.cy - sizClient.cy ;
    }

    if (sizClient.cx >= sizBitmap.cx) {
        WinEnableWindow(hwndHorz, FALSE) ;
    } else {
        diff = sizBitmap.cx - sizClient.cx ;
	WinSendMsg(hwndHorz, SBM_SETTHUMBSIZE,
	        MPFROM2SHORT(sizClient.cx, sizBitmap.cx), NULL) ;
        WinSendMsg(hwndHorz, SBM_SETSCROLLBAR,
	        MPFROM2SHORT(offClient.x, 0), MPFROM2SHORT(0, diff)) ;
        WinEnableWindow(hwndHorz, TRUE) ;
    }

    if (sizClient.cy >= sizBitmap.cy) {
        WinEnableWindow(hwndVert, FALSE) ;
    } else {
        diff = sizBitmap.cy - sizClient.cy ;
	vpos = diff - offClient.y ;
	WinSendMsg(hwndVert, SBM_SETTHUMBSIZE,
	        MPFROM2SHORT(sizClient.cy, sizBitmap.cy), NULL) ;
        WinSendMsg(hwndVert, SBM_SETSCROLLBAR,
	        MPFROM2SHORT(vpos, 0), MPFROM2SHORT(0, diff)) ;
        WinEnableWindow(hwndVert, TRUE) ;
    }
}

/*
 * bmpMapHorz - adjust horizontal mapping
 */

#define PAGEDIV 8

void    bmpMapHorz(USHORT cmd, SHORT x)
{
    SHORT       new, lim ;
    BOOL        set ;
    
    lim = sizBitmap.cx - sizClient.cx ;
    
    switch (cmd) {
    case SB_SLIDERTRACK    :
    case SB_SLIDERPOSITION :
        new = x     ;
	set = FALSE ;
	break ;
    case SB_LINELEFT :
        new = offClient.x - 1 ;
	set = TRUE ;
	break ;
    case SB_LINERIGHT :
        new = offClient.x + 1 ;
	set = TRUE ;
	break ;
    case SB_PAGELEFT :
        new = offClient.x - (sizBitmap.cx / PAGEDIV) ;
	set = TRUE ;
	break ;
    case SB_PAGERIGHT :
        new = offClient.x + (sizBitmap.cx / PAGEDIV) ;
	set = TRUE ;
	break ;
    default :
        return ;
    }
    
    if (new < 0)   new = 0   ;
    if (new > lim) new = lim ;
    
    if (set) {
        WinSendMsg(hwndHorz, SBM_SETSCROLLBAR,
	        MPFROM2SHORT(new, 0), MPFROM2SHORT(0, lim)) ;
    }
    if (new != offClient.x) {
        offClient.x = new ;
        WinInvalidateRect(hwndClient, NULL, FALSE) ;
    }
}

/*
 * bmpMapVert - adjust vertical mapping
 */

void    bmpMapVert(USHORT cmd, SHORT y)
{
    SHORT       new, lim ;
    BOOL        set ;
    
    lim = sizBitmap.cy - sizClient.cy ;
    
    switch (cmd) {
    case SB_SLIDERTRACK    :
    case SB_SLIDERPOSITION :
        new = lim - y ;
	set = FALSE ;
	break ;
    case SB_LINEUP :
        new = offClient.y + 1 ;
	set = TRUE ;
	break ;
    case SB_LINEDOWN :
        new = offClient.y - 1 ;
	set = TRUE ;
	break ;
    case SB_PAGEUP :
        new = offClient.y + (sizBitmap.cy / PAGEDIV) ;
	set = TRUE ;
	break ;
    case SB_PAGEDOWN :
        new = offClient.y - (sizBitmap.cy / PAGEDIV) ;
	set = TRUE ;
	break ;
    default :
        return ;
    }

    if (new < 0)   new = 0   ;
    if (new > lim) new = lim ;
    
    if (set) {
        WinSendMsg(hwndVert, SBM_SETSCROLLBAR,
	        MPFROM2SHORT((lim - new), 0), MPFROM2SHORT(0, lim)) ;
    }
    if (new != offClient.y) {
        offClient.y = new ;
	WinInvalidateRect(hwndClient, NULL, FALSE) ;
    }
}

/*
 * bmpMapPoint - convert coordinates between window/bitmap
 */

void    bmpMapPoint(PPOINTL pt, BOOL toBmp)
{
    if (toBmp) {
        pt->x += offClient.x ;
	pt->y += offClient.y ;
    } else {
        pt->x -= offClient.x ;
	pt->y -= offClient.y ;
    }
}

/*
 * bmpPaint - draw a region of bitmap to client window
 */

void    bmpPaint(HPS hps, PRECTL rct)
{
    POINTL  apt[3] ;
    
    apt[0].x = rct->xLeft   ;
    apt[0].y = rct->yBottom ;
    apt[1].x = rct->xRight  ;
    apt[1].y = rct->yTop    ;
    apt[2].x = offClient.x + rct->xLeft   ;
    apt[2].y = offClient.y + rct->yBottom ;

    GpiBitBlt(hps, hpsBitmap, 3, apt, ROP_SRCCOPY, 0) ;
}

/*
 * bmpUpdate - update rectangle (map bitmap to window)
 */

void    bmpUpdate(int x, int y, int w, int h)
{
    RECTL   rct ;
    
    rct.xLeft   = x     - offClient.x - 1 ;
    rct.xRight  = x + w - offClient.x + 1 ;
    rct.yBottom = y     - offClient.y - 1 ;
    rct.yTop    = y + h - offClient.y + 1 ;

    WinInvalidateRect(hwndClient, &rct, FALSE) ;
}

/*
 * bmpRectDraw - raw rectangle, draw image data
 */

void    bmpRectDraw(int x, int y, int w, int h, PUCHAR bits)
{
    BITMAPINFO2 bmi ;
    POINTL  apt[4] ;

    if (w == 0 || h == 0) {
        return ;
    }
    
    memset(&bmi, 0, sizeof(BITMAPINFO2)) ;
    bmi.cbFix = 16     ;
    bmi.cPlanes   = 1  ;
    bmi.cBitCount = 24 ;
    
    bmi.cx = w ;
    bmi.cy = h ;
    
    apt[0].x = x ;
    apt[0].y = y ;
    apt[1].x = x + w - 1 ;
    apt[1].y = y + h - 1 ;
    apt[2].x = 0 ;
    apt[2].y = 0 ;
    apt[3].x = w ;
    apt[3].y = h ;

    GpiDrawBits(hpsBitmap, bits, &bmi, 4, apt, ROP_SRCCOPY, 0) ;
}

/*
 * bmpRectCopy - copy rectangle
 */

void    bmpRectCopy(int x, int y, int w, int h, int sx, int sy)
{
    POINTL      apt[3] ;

    apt[0].x = x ;
    apt[0].y = y ;
    apt[1].x = x + w ;
    apt[1].y = y + h ;
    apt[2].x = sx ;
    apt[2].y = sy ;

    GpiBitBlt(hpsBitmap, hpsBitmap, 3, apt, ROP_SRCCOPY, 0) ;
}

/*
 * bmpRectColor - change color of (following) solid rectangles
 *      replaced with macro   see pmvncdef.h
 */

#if 0
void    bmpRectColor(ULONG color)
{
    GpiSetColor(hpsBitmap, color) ;
}
#endif

/*
 * bmpRectFill - solid fill
 */

void    bmpRectFill(int x, int y, int w, int h)
{
    POINTL      pt1, pt2 ;

    if (w == 0 || h == 0) {
        return ;
    }
    
    pt2.x = (pt1.x = x) + w - 1 ;
    pt2.y = (pt1.y = y) + h - 1 ;

    if (w > 1 && h > 1) {
        GpiMove(hpsBitmap, &pt1) ;
	GpiBox(hpsBitmap, DRO_OUTLINEFILL, &pt2, 0, 0) ;
    } else if (w == 1 || h == 1) {
        GpiMove(hpsBitmap, &pt1) ;
	GpiLine(hpsBitmap, &pt2) ;
    } else {
        GpiSetPel(hpsBitmap, &pt1) ;
    }
}
